"""The product block that describes a site subscription."""

import re
from typing import Annotated

from annotated_types import doc
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle, strEnum
from pydantic import AfterValidator, Field


class SiteTier(strEnum):
    """The tier of a site, ranging from 1 to 4.

    A lower value corresponds to a larger amount of installed equipment, and a higher cardinality of the connectivity to
    and from a site.
    """

    TIER1 = 1
    TIER2 = 2
    TIER3 = 3
    TIER4 = 4


def validate_latitude(v: float) -> float:
    """Validate a latitude coordinate."""
    regex = re.compile(r"^-?([1-8]?\d(\.\d+)?|90(\.0+)?)$")
    if not regex.match(str(v)):
        msg = "Invalid latitude coordinate. Valid examples: '40.7128', '-74.0060', '90', '-90', '0'."
        raise ValueError(msg)
    return v


def validate_longitude(v: float) -> float:
    """Validate a longitude coordinate."""
    regex = re.compile(r"^-?(180(\.0+)?|((1[0-7]\d)|([1-9]?\d))(\.\d+)?)$")
    if not regex.match(str(v)):
        msg = "Invalid longitude coordinate. Valid examples: '40.7128', '-74.0060', '180', '-180', '0'."
        raise ValueError(msg)

    return v


LatitudeCoordinate = Annotated[
    float,
    Field(
        ge=-90,
        le=90,
    ),
    AfterValidator(validate_latitude),
    doc(
        "A latitude coordinate, modeled as a string. "
        "The coordinate must match the format conforming to the latitude range of -90 to +90 degrees. "
        "It can be a floating-point number or an integer. Valid examples: 40.7128, -74.0060, 90, -90, 0."
    ),
]

LongitudeCoordinate = Annotated[
    float,
    Field(
        ge=-180,
        le=180,
    ),
    AfterValidator(validate_longitude),
    doc(
            "A longitude coordinate, modeled as a string. "
            "The coordinate must match the format conforming to the longitude "
            "range of -180 to +180 degrees. It can be a floating-point number or an integer. "
            "Valid examples: 40.7128, -74.0060, 180, -180, 0."
    ),
]


class SiteBlockInactive(
    ProductBlockModel,
    lifecycle=[SubscriptionLifecycle.INITIAL],
    product_block_name="SiteBlock",
):
    """A site that's currently inactive, see :class:`SiteBlock`."""

    site_name: str | None = None
    site_city: str | None = None
    site_country: str | None = None
    site_country_code: str | None = None
    site_latitude: LatitudeCoordinate | None = None
    site_longitude: LongitudeCoordinate | None = None
    site_internal_id: int | None = None
    site_bgp_community_id: int | None = None
    site_tier: SiteTier | None = None
    site_ts_address: str | None = None


class SiteBlockProvisioning(SiteBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    """A site that's currently being provisioned, see :class:`SiteBlock`."""

    site_name: str
    site_city: str
    site_country: str
    site_country_code: str
    site_latitude: LatitudeCoordinate
    site_longitude: LongitudeCoordinate
    site_internal_id: int
    site_bgp_community_id: int
    site_tier: SiteTier
    site_ts_address: str


class SiteBlock(SiteBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    """A site that's currently available for routers and services to be hosted at."""

    #:  The name of the site, that will dictate part of the :term:`FQDN` of routers that are hosted at this site. For
    #:  example: ``router.X.Y.geant.net``, where X denotes the name of the site.
    site_name: str
    #:  The city at which the site is located.
    site_city: str
    #:  The country in which the site is located.
    site_country: str
    #:  The code of the corresponding country. This is also used for the :term:`FQDN`, following the example given for
    #:  the site name, the country code would end up in the Y position.
    site_country_code: str
    #:  The latitude of the site, used for :term:`SNMP` purposes.
    site_latitude: LatitudeCoordinate
    #:  Similar to the latitude, the longitude of a site.
    site_longitude: LongitudeCoordinate
    #:  The internal ID used within GÉANT to denote a site.
    site_internal_id: int
    #:  The :term:`BGP` community ID of a site, used to advertise routes learned at this site.
    site_bgp_community_id: int
    #:  The tier of a site, as described in :class:`SiteTier`.
    site_tier: SiteTier
    #:  The address of the terminal server that this router is connected to. The terminal server provides out of band
    #:  access. This is required in case a link goes down, or when a router is initially added to the network and it
    #:  does not have any IP trunks connected to it.
    site_ts_address: str
