"""A creation workflow that adds an existing router to the service database."""

from orchestrator import workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, done, init, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from pydantic import ConfigDict

from gso.products import ProductName
from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import ImportedRouterInactive
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name, get_site_by_name
from gso.utils.helpers import generate_fqdn
from gso.utils.shared_enums import IPv4AddressType, IPv6AddressType, PortNumber, Vendor


@step("Create subscription")
def create_subscription(partner: str) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName.IMPORTED_ROUTER)
    subscription = ImportedRouterInactive.from_product_id(product_id, partner_id)

    return {
        "subscription": subscription,
        "subscription_id": subscription.subscription_id,
    }


def initial_input_form_generator() -> FormGenerator:
    """Generate a form that is filled in using information passed through the :term:`API` endpoint."""

    class ImportRouter(FormPage):
        model_config = ConfigDict(title="Import Router")

        partner: str
        router_site: str
        hostname: str
        ts_port: int
        router_vendor: Vendor
        router_role: RouterRole
        router_lo_ipv4_address: IPv4AddressType
        router_lo_ipv6_address: IPv6AddressType
        router_lo_iso_address: str

    user_input = yield ImportRouter

    return user_input.dict()


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedRouterInactive,
    hostname: str,
    ts_port: PortNumber,
    router_site: str,
    router_role: RouterRole,
    router_vendor: Vendor,
    router_lo_ipv4_address: IPv4AddressType | None = None,
    router_lo_ipv6_address: IPv6AddressType | None = None,
    router_lo_iso_address: str | None = None,
) -> State:
    """Initialise the router subscription using input data."""
    subscription.router.router_ts_port = ts_port
    router_site_obj = get_site_by_name(router_site).site
    subscription.router.router_site = router_site_obj
    fqdn = generate_fqdn(hostname, router_site_obj.site_name, router_site_obj.site_country_code)
    subscription.router.router_fqdn = fqdn
    subscription.router.router_role = router_role
    subscription.router.router_access_via_ts = False
    subscription.description = f"Router {fqdn}"
    subscription.router.router_lo_ipv4_address = router_lo_ipv4_address
    subscription.router.router_lo_ipv6_address = router_lo_ipv6_address
    subscription.router.router_lo_iso_address = router_lo_iso_address
    subscription.router.vendor = router_vendor

    return {"subscription": subscription}


@workflow(
    "Import router",
    initial_input_form=initial_input_form_generator,
    target=Target.SYSTEM,
)
def create_imported_router() -> StepList:
    """Import a router without provisioning it."""
    return (
        init
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
